

#ifdef WIN32
#include <windows.h>
#include <GL/gl.h>
#endif

#include "vgev.h"
#include "vgev_data.h"


/* ***************************************************************************
*************************************************************************** */
void drawable_manager::resort() {

   drawable_symbol *symbol, *ptr;
   dbl_llist_manager tman, hman;
   
   symbol = (drawable_symbol *)head;
   head = tail = NULL;
   dest();
   
   while (symbol) {
      ptr = symbol;
      symbol = (drawable_symbol *)symbol->next;

      if (ptr->flags & DRAWABLEFLAG_PICK) {
         tman.append(ptr, NULL);
         continue;
      }

      if (ptr->flags & DRAWABLEFLAG_HIGHLIGHT) {
         hman.append(ptr, NULL);
         continue;
      }

      append(ptr, NULL);
   }

   symbol = (drawable_symbol *)hman.head;
   hman.head = hman.tail = NULL;
   hman.dest();
   
   while (symbol) {
      ptr = symbol;
      symbol = (drawable_symbol *)symbol->next;

      append(ptr, NULL); 
   }
   
   symbol = (drawable_symbol *)tman.head;
   tman.head = tman.tail = NULL;
   tman.dest();
   
   while (symbol) {
      ptr = symbol;
      symbol = (drawable_symbol *)symbol->next;

      append(ptr, NULL); 
   }
   
}


/* ****************************************************************************
**************************************************************************** */
void drawable_manager::recalc(canvas *drawable) {

   drawable_symbol *ptr;

   for (ptr = (drawable_symbol *)head; ptr; ptr = (drawable_symbol *)ptr->next)
      ptr->recalc(drawable);

   flags &= ~SYMBOLFLAG_RECALC;
}


/* ****************************************************************************
**************************************************************************** */
void drawable_manager::set_pick(drawable_symbol *symbol, canvas *drawable) {

   drawable_symbol *ptr;

   for (ptr = (drawable_symbol *)head; ptr; ptr = (drawable_symbol *)ptr->next)
      ptr->set_pick(symbol == ptr, drawable);
}


/* ****************************************************************************
**************************************************************************** */
drawable_symbol *drawable_manager::find_object(int bx, int by, drawable_symbol **pick, unsigned int mask) {

   drawable_symbol *old;
   double  dx,dy;
   double pickdist2 = PICK_RADIUS * PICK_RADIUS;
   drawable_symbol *ptr, *qtr;
   
   old = *pick;
   *pick = NULL;

   ptr = old;
   qtr = NULL;

   if (ptr)
      for (ptr = qtr = (drawable_symbol *)ptr->next; ptr; ptr=(drawable_symbol *)ptr->next) {

         if (!(ptr->flags & mask))
            continue;

         dx = ptr->x - bx;
         dy = ptr->y - by;

         if (dx*dx+dy*dy <= pickdist2)
            return *pick = ptr;
      }

   for (ptr=(drawable_symbol *)head; ptr != qtr; ptr=(drawable_symbol *)ptr->next) {
	   
      if (!(ptr->flags & mask))
         continue;

      dx = ptr->x - bx;
      dy = ptr->y - by;

      if (dx*dx+dy*dy <= pickdist2)
         return *pick = ptr;
   }

   return NULL;   
}


/* ****************************************************************************
**************************************************************************** */
drawable_symbol::drawable_symbol() {

   flags = DRAWABLEFLAG_NULL;
   source = NULL;
   icon = NULL;
   lat = lon = 0;
   x = y = 0;
   alt = 0;
   agl_flag = 0;
}


/* ****************************************************************************
**************************************************************************** */
void drawable_symbol::render(canvas *tgev) {

   if ((flags & DRAWABLEFLAG_VISIBLE) && icon)
      icon->render(tgev, color, x, y);
}

/* ****************************************************************************
**************************************************************************** */
void drawable_symbol::set_pick(int pickflag, canvas *drawable) {

   if (pickflag)
      flags |= DRAWABLEFLAG_PICK;
   else
      flags &= ~DRAWABLEFLAG_PICK;

   drawable->redraw_flag = 1;
}


/* ****************************************************************************
**************************************************************************** */
void drawable_symbol::recalc(canvas *drawable) {

   if (drawable->clip(lon, lat, &x, &y))
      flags &= ~DRAWABLEFLAG_VISIBLE;
   else
      flags |= DRAWABLEFLAG_VISIBLE;

   drawable->redraw_flag = 1;
}


/* **************************************************************
************************************************************** */
void overlay_symbol::render(canvas *tgev) {

   if (!(flags & DRAWABLEFLAG_VISIBLE))
      return;
      
   glPointSize((flags & DRAWABLEFLAG_PICK) ? 8.0f : 4.0f);
   tgev->tpColor3ubv(color, 0);

   glBegin(GL_POINTS);
      glVertex2i(x, y);
   glEnd();

   glPointSize(1.0);
}


/* **************************************************************
************************************************************** */
overlay_point::overlay_point() {
	
   symbol = new overlay_symbol;
   symbol->source = this;
   symbol->flags |= DRAWABLEFLAG_OVERLAY;
   symbol_manager.insert(symbol, NULL);
}

	  
/* **************************************************************
************************************************************** */
overlay_point::~overlay_point() {
	
   symbol_manager.remove(symbol);
   delete symbol;
}




/* ****************************************************************************
**************************************************************************** */
void platform_symbol::render(canvas *tgev) {

   vector4uc c;
   
   if (!(flags & DRAWABLEFLAG_VISIBLE) || !icon)
      return;

   // object
   if (flags & DRAWABLEFLAG_PICK) {
      c[0] = 255;
      c[1] = 0;
      c[2] = 0;
   }
   
   else if (flags & DRAWABLEFLAG_HIGHLIGHT) {
      c[0] = 255;
      c[1] = 255;
      c[2] = 0;
   }

   else
      copyarray3(c, color);

   // heading vector
   tgev->tpColor3ubv(c, 1);
   glBegin(GL_LINES);
      glVertex2i(x, y);
      glVertex2d(x-16.0*sin(((VGEVplatform_data *)source)->orient[2]), y+16.0*cos(((VGEVplatform_data *)source)->orient[2]));
   glEnd();

   // name
   tgev->DrawString(x ,y+16, ((VGEVplatform_data *)source)->name.string,((VGEVplatform_data *)source)->name.stringlen());

   // object
   icon->render(NULL, NULL, x, y);
}


/* **************************************************************
************************************************************** */
int overlay_polyline::parse(FILE *infile, char *token) {

   overlay_point *ptr;

   overlay_primitive::parse(infile, token);

   if (fillflag) {
      get_token(infile, token);
      color[3] = atoi(token);
   }

   get_token(infile, token); // '{'

   while (get_token(infile, token) && token[0] != '}') {
      ptr = new overlay_point;
      list.append(ptr, NULL);

      ptr->symbol->lat = atof(token);

      get_token(infile, token);
      ptr->symbol->lon = atof(token);

      copyarray3(ptr->symbol->color, color);
   }

   return 1;
}


/* **************************************************************
************************************************************** */
int overlay_polyline::parse(overlay_node *node) {

   overlay_point *ptr;
   overlay_point_node *otr;
   
   overlay_polygon_node *x = (overlay_polygon_node *)node;
   
   overlay_primitive::parse(node);

   fillflag = x->fillflag;

   for (otr = (overlay_point_node *)((overlay_polygon_node *)node)->point_list.head; otr; otr = (overlay_point_node *)otr->next) {
      ptr = new overlay_point;

      ptr->symbol->lat = otr->latlon[0];
      ptr->symbol->lon = otr->latlon[1];

      copyarray3(ptr->symbol->color, color);
   }
      
   return 1;
}


/* **************************************************************
************************************************************** */
void overlay_polyline::render(canvas *tgev) {

   overlay_point *ptr;

   if (fillflag) {
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

      tgev->tpColor4ubv(color, 0);
      glBegin(GL_POLYGON);
   }

   else {
      tgev->tpColor3ubv(color, 0);
      glBegin(GL_LINE_STRIP);
   }

   for (ptr = (overlay_point *)list.head; ptr; ptr = (overlay_point *)ptr->next)
      glVertex2i(ptr->symbol->x, ptr->symbol->y);

   glEnd();

   if (fillflag)
      glDisable(GL_BLEND);
}


/* **************************************************************
************************************************************** */
int overlay_primitive::parse(FILE *infile, char *token) {

   get_token(infile, token);
   color[0] = atoi(token);

   get_token(infile, token);
   color[1] = atoi(token);

   get_token(infile, token);
   color[2] = atoi(token);

   color[3] = 255;

   return 1;
}


/* **************************************************************
************************************************************** */
int overlay_primitive::parse(overlay_node *node) {

   id = node->id;
   copyarray4(color, node->color);
   return 1;
}

